#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <climits>
#include <cstdio>
#include <iostream>
#include <string>
#include <vector>

extern "C" char **environ;

using namespace std;

vector<string> split(char const *str);
string join(vector<string> const &in, unsigned to);
bool readable(string const &file);
bool ownedbyme(string const &file);

enum Shell {
	BASH,
	TCSH
};

int main(int argc, char **argv) {
bool verbose = false;
Shell shell = BASH;
char *olddir = argv[1];
struct stat info;

	if (getuid() == 0) {
		cerr << argv[0] << ": error, refusing to run as root" << endl;
		return 1;
	}
		
	// parse options
	{
	int i;
		for (i = 1; i < argc && argv[i][0] == '-'; ++i)
			switch (argv[i][1]) {
				case 'v' : // verbose
					verbose = true;
				break;
				case 't' : // tcsh
					shell = TCSH;
				break;
				case 'h' :
					cout << "usage: " << argv[0] << " [-v] [-t] [-h] <old directory>" << endl
						<< "  -v    verbose" << endl
						<< "  -t    display TCSH commands rather than BASH" << endl
						<< "  -h    help" << endl;
					return 0;
				break;
			}
		if (i >= argc) {
			cerr << argv[0] << ": error, path must be specified" << endl;
			return 1;
		}
		olddir = argv[i];
	}

	// get base working directories
vector<string> last_dir, current_dir;

	{
	char tmp[strlen(olddir) + 1];
		strcpy(tmp, olddir);
		if (tmp[0] != '/') {
			cerr << argv[0] << ": error, path must be absolute" << endl;
			return 1;
		}
		while (tmp[strlen(tmp) - 1] == '/') tmp[strlen(tmp) - 1] = 0;
		last_dir = split(tmp);
	}
	{
	char tmp[PATH_MAX];
		getcwd(tmp, PATH_MAX);
		while (tmp[strlen(tmp) - 1] == '/') tmp[strlen(tmp) - 1] = 0;
		current_dir = split(tmp);
	}

unsigned common_root;

	// Find common root path
	for (common_root = 0; last_dir.size() > common_root && current_dir.size() > common_root && last_dir[common_root] == current_dir[common_root] ; common_root++) ;
	if (common_root) common_root--;

	// Move up to common root from old path
	for (unsigned i = last_dir.size() - 1; i > common_root; i--) {
	string leaving = join(last_dir, i);

		if (readable(leaving + "/.onleave")) {
			if (!ownedbyme(leaving + "/.onleave"))
				cerr << argv[0] << ": warning, not executing \"" << leaving << "/.onleave\" as i do not own it" << endl;
			else {
				if (verbose) cerr << "ondir: called .onleave in '" << leaving << "'" << endl;
				if (shell == TCSH) cout << "setenv ";
				cout << "ONDIRWD=\"" << leaving << "\" ";
				if (shell == BASH)
					cout << ".";
				if (shell == TCSH)
					cout << "; source";
				cout << " \"" << leaving << "/.onleave\";" << endl;
			}
		}
	}

	// Move in to new path from common root
	for (unsigned i = common_root + 1; i < current_dir.size(); i++) {
	string entering = join(current_dir, i);

		if (readable(entering + "/.onenter")) {
			if (!ownedbyme(entering + "/.onenter"))
				cerr << argv[0] << ": warning, not executing \"" << entering << "/.onenter\" as i do not own it" << endl;
			else {
				if (verbose) cerr << "ondir: called .onenter in '" << entering << "'" << endl;
				if (shell == TCSH) cout << "setenv ";
				cout << "ONDIRWD=\"" << entering << "\" ";
				if (shell == BASH)
					cout << ".";
				if (shell == TCSH)
					cout << "; source";
				cout << " \"" << entering << "/.onenter\";" << endl;
			}
		}
	}
	return 0;
}

/* Split a path into a vector */
vector<string> split(char const *str) {
vector<string> out;
char const *last;

	if (str[0] == '/') str++;
	for (last = str; *str; str++) {
		if (*str == '/') {
			out.push_back(string(last, str - last));
			last = str + 1;
		}
	}
	out.push_back(string(last, str - last));
	return out;
}

/* Join a split path into a string */
string join(vector<string> const &in, unsigned to) {
string out;

	for (unsigned i = 0; i <= to; ++i)
		out += "/" + in[i];
	return out;
}

/* Return true if file is readable */
bool readable(string const &file) {
struct stat s;

	if (!stat(file.c_str(), &s) && 
		s.st_mode & (S_IRUSR | S_IRGRP | S_IROTH))
			return true;
	return false;
}

/* Return true if file is readable */
bool ownedbyme(string const &file) {
struct stat s;

	if (!stat(file.c_str(), &s) && s.st_uid == getuid())
			return true;
	return false;
}
